// @(#)root/minuit2:$Id$
// Authors: M. Winkler, F. James, L. Moneta, A. Zsenei   2003-2005

/**********************************************************************
 *                                                                    *
 * Copyright (c) 2005 LCG ROOT Math team,  CERN/PH-SFT                *
 *                                                                    *
 **********************************************************************/

#ifndef ROOT_Minuit2_MnMatrixfwd
#define ROOT_Minuit2_MnMatrixfwd


namespace ROOT {

namespace Minuit2 {

class LASymMatrix;
class LAVector;

typedef LASymMatrix MnAlgebraicSymMatrix;
typedef LAVector MnAlgebraicVector;

} // namespace Minuit2

} // namespace ROOT

#endif // ROOT_Minuit2_MnMatrixfwd
