// @(#)root/minuit2:$Id$
// Authors: M. Winkler, F. James, L. Moneta, A. Zsenei   2003-2005

/**********************************************************************
 *                                                                    *
 * Copyright (c) 2005 LCG ROOT Math team,  CERN/PH-SFT                *
 *                                                                    *
 **********************************************************************/

#ifndef ROOT_Minuit2_MnConfig
#define ROOT_Minuit2_MnConfig

#ifdef _MSC_VER
#pragma warning(disable : 4244) // conversion from __w64 to int
#endif

#if defined(__sun) && !defined(linux)
#include <cstdlib>
#endif

#endif
